@@ -1,9 +1,9 @@
 class RDG(object):
-    def __init__(self, locus_stop=1000, nodes=None, edges=None):
+    def __init__(self, locus_stop=1000, nodes=None, edges=None, name=None):
         """
         initialise a fresh empty RDG 
         """
-        self.locus = ""
+        self.locus = name
         self.locus_start = 0
         self.locus_stop = locus_stop
 
@@ -15,32 +15,27 @@ class RDG(object):
         transcript_length = locus_stop - self.locus_start
 
         nodes = {
-                1:Node(key=1, node_type="5_prime", coordinates=(0,1), edges_out=[1], nodes_out=[2]),
-                2:Node(key=2, node_type="3_prime", coordinates=(transcript_length -1, transcript_length), edges_in=[1], nodes_in=[1])
+                1:Node(key=1, node_type="5_prime", coordinates=0, edges_out=[1], nodes_out=[2]),
+                2:Node(key=2, node_type="3_prime", coordinates=transcript_length -1, edges_in=[1], nodes_in=[1])
             }
 
         edges = {1: Edge(1, "untranslated", from_node=1, to_node=2, coordinates=(1, transcript_length -1))}
 
 
-        self.node_count = 2 
-        self.edge_count = 1 
         self.edges = edges
         self.nodes = nodes
 
-    def load(self, locus_name="unnamed", locus_start=0, locus_stop=1000, nodes=None, edges=None):
+
+    def Load(self, locus_name="unnamed", locus_start=0, locus_stop=1000, nodes=None, edges=None):
         """
         create a RDG from the given paramaters 
         """
         if nodes:
             self.nodes = nodes
-            self.node_count = len(nodes.keys())
-
 
         if edges:
             self.edges = edges 
-            self.edge_count = len(edges.keys())
         
-
         self.locus = locus_name
         self.locus_start = locus_start
         self.locus_stop = locus_stop
@@ -53,11 +48,11 @@ class RDG(object):
         orf coordinates = 10, 100
         '''
         nodes = {
-            1:Node(key=1, node_type="5_prime", coordinates=(0,1), edges_out=[1], nodes_out=[3]),
-            2:Node(key=2, node_type="3_prime", coordinates=(1000 -1, 1000), edges_in=[2], nodes_in=[3]),
-            3:Node(key=3, node_type="start", coordinates=(10,11), edges_in=[1], edges_out=[2,3], nodes_in=[1], nodes_out=[2,4]),
-            4:Node(key=4, node_type="stop", coordinates=(100,101), edges_in=[3], edges_out=[4], nodes_in=[3], nodes_out=[5]),
-            5:Node(key=5, node_type="3_prime", coordinates=(1000 -1, 1000), edges_in=[4], nodes_in=[4])
+            1:Node(key=1, node_type="5_prime", coordinates=0, edges_out=[1], nodes_out=[3]),
+            2:Node(key=2, node_type="3_prime", coordinates=1000 -1, edges_in=[2], nodes_in=[3]),
+            3:Node(key=3, node_type="start", coordinates=10, edges_in=[1], edges_out=[2,3], nodes_in=[1], nodes_out=[2,4]),
+            4:Node(key=4, node_type="stop", coordinates=100, edges_in=[3], edges_out=[4], nodes_in=[3], nodes_out=[5]),
+            5:Node(key=5, node_type="3_prime", coordinates=1000 -1, edges_in=[4], nodes_in=[4])
             }
         edges = {
             1:Edge(1, "untranslated", from_node=1, to_node=3, coordinates=(1, 10 -1)),
@@ -66,7 +61,7 @@ class RDG(object):
             4:Edge(4, "untranslated", from_node=4, to_node=5, coordinates=(101, 1000 -1))
         }
         g = RDG()
-        g = RDG.load(g, nodes=nodes, edges=edges)
+        g = RDG.Load(g, nodes=nodes, edges=edges)
         return g
 
 
@@ -88,9 +83,9 @@ class RDG(object):
         '''
         return a list of edges (from, to) in the graph
         '''
-        from_to = [] 
+        from_to = {} 
         for edge in self.edges.keys():
-            from_to.append((self.edges[edge].from_node, self.edges[edge].to_node))
+            from_to[(self.edges[edge].from_node, self.edges[edge].to_node)] = edge
 
         return from_to
 
@@ -116,6 +111,17 @@ class RDG(object):
         else:
             return 1
 
+    def get_key_from_position(self, position, node_type):
+        '''
+        return the node key for the node of specified type at the stated position 
+        '''
+        for node in self.nodes:
+            if self.nodes[node].node_start == position and self.nodes[node].node_type == node_type:
+                return node
+        raise ValueError(f"There is no stop codon at the locus ({position}) specified for readthrough")
+
+
+
     def remove_edge(self, edge_key):
         '''
         remove the edge with the given key from the graph Also removing references from adjacent nodes 
@@ -149,6 +155,8 @@ class RDG(object):
         self.nodes[to_node_key].input_nodes.append(from_node_key)
 
         self.edges[edge.key] = edge 
+        self.edges[edge.key].frame = self.nodes[from_node_key].node_start%3
+
 
 
     def add_node(self, node):
@@ -200,8 +208,8 @@ class RDG(object):
         handle inserting node into just one edge including down path 
         '''
         five_prime_edge_coords = (edge.coordinates[0], start_node.node_start - 1)
-        coding_edge_coords = (start_node.node_start, stop_node.node_stop)
-        three_prime_edge_coords = (stop_node.node_stop + 1, edge.coordinates[1])
+        coding_edge_coords = (start_node.node_start, stop_node.node_start)
+        three_prime_edge_coords = (stop_node.node_start + 1, edge.coordinates[1])
 
         edge_key = self.get_new_edge_key()
         five_prime = Edge(key=edge_key, edge_type="untranslated", from_node=edge.from_node, to_node=start_node.key, coordinates=five_prime_edge_coords)
@@ -217,7 +225,7 @@ class RDG(object):
 
 
         terminal_node_key = self.get_new_node_key()
-        terminal_node = Node(key=terminal_node_key, node_type="3_prime", coordinates=(edge.coordinates[1],edge.coordinates[1] + 1), edges_in=[], edges_out=[], nodes_in=[], nodes_out=[])
+        terminal_node = Node(key=terminal_node_key, node_type="3_prime", coordinates=edge.coordinates[1], edges_in=[], edges_out=[], nodes_in=[], nodes_out=[])
 
         edge_key = self.get_new_edge_key()
         three_prime = Edge(key=edge_key, edge_type="untranslated", from_node=stop_node.key, to_node=terminal_node.key, coordinates=three_prime_edge_coords)
@@ -225,7 +233,7 @@ class RDG(object):
         self.add_node(terminal_node)
         self.add_edge(three_prime, stop_node.key, terminal_node.key)
 
-        self.update_edge(edge.key, start_node.key, edge.to_node, (start_node.node_stop, self.nodes[edge.to_node].node_start))
+        self.update_edge(edge.key, start_node.key, edge.to_node, (start_node.node_start, self.nodes[edge.to_node].node_start))
 
 
     def is_input_edge_translated(self, node):
@@ -262,6 +270,10 @@ class RDG(object):
         '''
         return path from root to given node as a list of edge keys
         '''
+        if self.nodes[node].input_nodes == []:
+            reached_root = True
+            return []
+
         path = [self.nodes[node].input_edges[0]]
         reached_root = False
 
@@ -272,28 +284,30 @@ class RDG(object):
             else:
                 path.append(self.nodes[in_node].input_edges[0])
                 node = in_node
-
         return path[::-1]
             
 
-    def check_translation_upstream(self, from_node):
+    def check_translation_upstream(self, from_node, upstream_limit=1):
         '''
         look upstream of an edges from node and see if any edges are of type 'translated' 
-        used in reinit functionality 
+        used in reinit functionality. Upstream limit refers to the number of ORFs allowed upstream. ie. can the be multiple reinitiation events or just one etc
         '''
         edge_path_to_node_from_root = self.root_to_node_of_acyclic_edge_path(from_node)
-
+        number_of_translated_regions = 0
         for edge in edge_path_to_node_from_root:
             if self.edges[edge].edge_type == 'translated':
-                return True
-        return False
+                number_of_translated_regions += 1
 
-    def add_open_reading_frame(self, start_codon_position, stop_codon_position, reinitiation=False):
+        if number_of_translated_regions <= upstream_limit:
+            return False
+        return True
+
+
+    def add_open_reading_frame(self, start_codon_position, stop_codon_position, reinitiation=False, upstream_limit=0):
         '''
         Handles all operations related to adding a new decision to the graph. Includes objects in graph and corrects all objects involved
         '''
         exisiting_edges = self.get_edges()
-
         clashing_edges = [] 
         for edge in exisiting_edges:
             if start_codon_position in range(self.edges[edge].coordinates[0], self.edges[edge].coordinates[1]) and self.edges[edge].edge_type != "translated":
@@ -303,14 +317,131 @@ class RDG(object):
         
         for edge, upstream_node in clashing_edges:
             node_key = self.get_new_node_key()
-            start_node = Node(key=node_key, node_type="start", coordinates=(start_codon_position, start_codon_position + 2), edges_in=[], edges_out=[], nodes_in=[], nodes_out=[])
+            start_node = Node(key=node_key, node_type="start", coordinates=start_codon_position, edges_in=[], edges_out=[], nodes_in=[], nodes_out=[])
 
             node_key = node_key + 1
-            stop_node = Node(key=node_key, node_type="stop", coordinates=(stop_codon_position, stop_codon_position + 2), edges_in=[], edges_out=[], nodes_in=[], nodes_out=[])
-            if reinitiation or not self.check_translation_upstream(upstream_node):
-
+            stop_node = Node(key=node_key, node_type="stop", coordinates=stop_codon_position, edges_in=[], edges_out=[], nodes_in=[], nodes_out=[])
+            if reinitiation or not self.check_translation_upstream(upstream_node, upstream_limit=upstream_limit):
                 self.insert_ORF(self.edges[edge], start_node, stop_node)
+
     
+    def add_stop_codon_readthrough(self, readthrough_codon_position, next_stop_codon_position):
+        '''
+        Handles all operations related to adding a new decision to the graph at a stop codon. Includes objects in graph and corrects all objects involved
+
+        '''
+        readthrough_codon_key = self.get_key_from_position(readthrough_codon_position, "stop")
+        three_prime_terminal_key = self.nodes[readthrough_codon_key].output_nodes[0]
+
+        node_key = self.get_new_node_key()
+        terminal_node_key = node_key + 1
+
+        edge_key = self.get_new_edge_key()
+        edge_key2 = edge_key + 1
+
+        new_stop_node = Node(key=node_key, node_type="stop", coordinates=next_stop_codon_position, edges_in=[edge_key], edges_out=[edge_key2], nodes_in=[readthrough_codon_key], nodes_out=[terminal_node_key])
+        self.add_node(new_stop_node)
+
+        coding = Edge(key=edge_key, edge_type="translated", from_node=readthrough_codon_key, to_node=new_stop_node.key, coordinates=(self.nodes[readthrough_codon_key].node_start, next_stop_codon_position))
+        self.add_edge(coding, readthrough_codon_key, new_stop_node.key)
+
+        self.nodes[readthrough_codon_key].output_edges.append(coding.key)
+        self.nodes[readthrough_codon_key].output_nodes.append(new_stop_node.key)
+        self.nodes[readthrough_codon_key].node_type = "readthrough_stop"
+        
+        three_prime = Edge(key=edge_key2, edge_type="untranslated", from_node=new_stop_node.key, to_node=terminal_node_key, coordinates=(new_stop_node.node_start, self.nodes[three_prime_terminal_key].node_start))
+
+        terminal_node_key = self.get_new_node_key()
+        terminal_node = Node(key=terminal_node_key, node_type="3_prime", coordinates=self.nodes[three_prime_terminal_key].node_start, edges_in=[edge_key2], edges_out=[], nodes_in=[new_stop_node.key], nodes_out=[])
+        self.add_node(terminal_node)
+
+        self.add_edge(three_prime, new_stop_node.key, terminal_node_key)
+
+
+    def add_frameshift(self, fs_position, next_stop_codon_position, shift):
+        '''
+        add a frameshifting event 
+        '''
+        exisiting_edges = self.get_edges()
+        clashing_edges = [] 
+        for edge in exisiting_edges:
+            if fs_position in range(self.edges[edge].coordinates[0], self.edges[edge].coordinates[1]) and self.edges[edge].edge_type == "translated":
+                upstream_node = self.edges[edge].from_node
+                clashing_edges.append((edge, upstream_node))  
+
+        for edge, upstream_node in clashing_edges:
+            # edge and upstrream are keys
+            shift_node_key = self.get_new_node_key()
+            new_stop_node_key = shift_node_key + 1
+            old_stop_node_key = self.edges[edge].to_node
+            new_terminal_key = new_stop_node_key + 1
+
+            old_stop_edge_key = self.get_new_edge_key()
+            new_stop_edge_key = old_stop_edge_key + 1
+            new_three_prime_edge_key = new_stop_edge_key + 1
+
+            shift_node = Node(key=shift_node_key, 
+            node_type="frameshift", 
+            coordinates=fs_position, 
+            edges_in=[edge], 
+            edges_out=[old_stop_edge_key, new_stop_edge_key], 
+            nodes_in=[upstream_node], 
+            nodes_out=[old_stop_node_key, new_stop_node_key])
+            self.add_node(shift_node)
+
+            self.nodes[upstream_node].output_nodes.remove(old_stop_node_key)
+            self.nodes[upstream_node].output_nodes.append(shift_node.key)
+
+            self.edges[edge].to_node = shift_node_key
+
+            self.nodes[old_stop_node_key].input_edges.remove(edge)
+            self.nodes[old_stop_node_key].input_edges.append(old_stop_edge_key)
+            self.nodes[old_stop_node_key].input_nodes.remove(upstream_node)
+            self.nodes[old_stop_node_key].input_nodes.append(shift_node_key)
+
+            
+            new_stop_node = Node(key=new_stop_node_key, 
+            node_type="stop", 
+            coordinates=next_stop_codon_position, 
+            edges_in=[new_stop_edge_key], 
+            edges_out=[new_three_prime_edge_key], 
+            nodes_in=[shift_node_key], 
+            nodes_out=[new_terminal_key])
+            self.add_node(new_stop_node)
+            
+            new_terminal_node = Node(key=new_terminal_key, 
+            node_type="3_prime",
+            coordinates=self.locus_stop, 
+            edges_in=[new_three_prime_edge_key], 
+            edges_out=[], 
+            nodes_in=[new_stop_node_key], 
+            nodes_out=[])
+            self.add_node(new_terminal_node)
+
+            old_stop_edge = Edge(key=old_stop_edge_key,
+            edge_type="translated",
+            from_node=shift_node_key,
+            to_node=old_stop_node_key,
+            coordinates=(shift_node.node_start, self.nodes[old_stop_node_key].node_start))
+            self.add_edge(old_stop_edge, from_node_key=shift_node_key, to_node_key=old_stop_node_key)
+            self.edges[old_stop_edge_key].frame = self.edges[edge].frame
+
+            new_stop_edge = Edge(key=new_stop_edge_key,
+            edge_type="translated",
+            from_node=shift_node_key,
+            to_node=new_stop_node_key,
+            coordinates=(shift_node.node_start, self.nodes[new_stop_node_key].node_start))
+            self.add_edge(new_stop_edge, from_node_key=shift_node_key, to_node_key=new_stop_node_key)
+
+
+            new_three_prime_edge = Edge(key=new_three_prime_edge_key,
+            edge_type="untranslated",
+            from_node=new_stop_node_key,
+            to_node=new_terminal_key,
+            coordinates=(new_stop_node.node_start, new_terminal_node.node_start))
+            self.add_edge(new_three_prime_edge, from_node_key=new_stop_node_key, to_node_key=new_terminal_key)
+
+
 
     def get_branch_points(self):
         '''
@@ -333,6 +464,7 @@ class RDG(object):
                 endpoints.append(node)
         return endpoints
 
+
     def get_startpoints(self):
         '''
         return list of start node keys 
@@ -354,17 +486,65 @@ class RDG(object):
                 translation_starts.append(node)
         return translation_starts
 
+
     def get_stop_nodes(self):
         '''
         return translation start keys
         '''
         translation_stops = [] 
         for node in self.nodes:
-            if self.nodes[node].node_type == "stop":
+            if "stop" in  self.nodes[node].node_type:
                 translation_stops.append(node)
         return translation_stops
 
+    
+    def get_orfs(self):
+        '''
+        return a list of coordinates (start, stop) describing the ORFs in the graph
+        ''' 
+        translation_starts = self.get_start_nodes()
 
+        orfs = []
+        for start in translation_starts:
+            downstream_nodes = self.nodes[start].output_nodes
+
+            for candidate_stop in downstream_nodes:
+                if self.nodes[candidate_stop].node_type == "stop":
+                    orfs.append((self.nodes[start].node_start, self.nodes[candidate_stop].node_start))
+
+                elif self.nodes[candidate_stop].node_type == "frameshift":
+                    # orfs.append((self.nodes[start].node_start, self.nodes[candidate_stop].node_start))
+                    fs_downstream_nodes = self.nodes[candidate_stop].output_nodes
+                    for new_candidate_stop in fs_downstream_nodes:
+                        if self.nodes[new_candidate_stop].node_type == "stop":
+                            orf = (self.nodes[start].node_start, self.nodes[candidate_stop].node_start, self.nodes[new_candidate_stop].node_start)
+                            if orf not in orfs:
+                                orfs.append(orf)
+
+                elif self.nodes[candidate_stop].node_type == "readthrough_stop":
+                    orfs.append((self.nodes[start].node_start, self.nodes[candidate_stop].node_start))
+                    readthrough_downstream_nodes = self.nodes[candidate_stop].output_nodes
+
+                    for new_candidate_stop in readthrough_downstream_nodes:
+                        if self.nodes[new_candidate_stop].node_type == "stop":
+                            orf = (self.nodes[start].node_start, self.nodes[new_candidate_stop].node_start)
+                            if orf not in orfs:
+                                orfs.append(orf)
+
+
+        
+        return orfs
+
+
+    def get_frameshifts(self):
+        '''
+        return a list of keys for all frameshift nodes
+        '''
+        frameshifts = [] 
+        for node in self.nodes:
+            if "frameshift" in  self.nodes[node].node_type:
+                frameshifts.append(node)
+        return frameshifts
 
 
     def print_paths(self):
@@ -388,29 +568,22 @@ class RDG(object):
         print("all paths", all_paths)
 
 
-
-
     def statistics(self):
         stats = {}
-        nodes = list(self.graph.keys())
+        nodes = list(self.nodes.keys())
         calc_dict = {
             "edges": [],
             "frames": [],
             "types": [],
-            "heights": [],
             "frames_freq": {},
             "types_freq": {},
-            "heights_freq": {},
         }
+        
+        calc_dict["edges"] = self.get_edges_from_to().keys()
 
         for i in nodes:
-            for j in self.graph[i].connected:
-                if (j, i) not in calc_dict["edges"]:
-                    calc_dict["edges"].append((i, j))
-
-            calc_dict["frames"].append(self.graph[i].frame)
-            calc_dict["types"].append(self.graph[i].node_type)
-            calc_dict["heights"].append(self.graph[i].height)
+            calc_dict["frames"].append(self.nodes[i].frame)
+            calc_dict["types"].append(self.nodes[i].node_type)
 
         for item in calc_dict["frames"]:
             if item in calc_dict["frames_freq"]:
@@ -424,12 +597,6 @@ class RDG(object):
             else:
                 calc_dict["types_freq"][item] = 1
 
-        for item in calc_dict["heights"]:
-            if item in calc_dict["heights_freq"]:
-                calc_dict["heights_freq"][item] += 1
-            else:
-                calc_dict["heights_freq"][item] = 1
-
         for item in calc_dict["types_freq"]:
             key = "Number_of_" + item
             stats[key] = calc_dict["types_freq"][item]
@@ -438,48 +605,27 @@ class RDG(object):
             key = "Number_of_nodes_in_frame_" + str(item)
             stats[key] = calc_dict["frames_freq"][item]
 
-        for item in calc_dict["heights_freq"]:
-            key = "Number_of_nodes_with_height_" + str(item)
-            stats[key] = calc_dict["heights_freq"][item]
 
         stats["Node_keys"] = nodes
-        stats["Number_of_nodes"] = self.node_count
+        stats["Number_of_nodes"] = len(self.nodes.keys())
         stats["Edges_keys"] = calc_dict["edges"]
         stats["Number_of_edges"] = len(calc_dict["edges"])
         stats["Number_of_node_types"] = len(calc_dict["types_freq"])
-
         return stats
 
+
     def describe(self):
         """
         print in the terminal a description of the graph
         """
-
+        output = ""
         stats = self.statistics()
         for entry in stats:
-            print(str(entry) + "\t" + str(stats[entry]))
-    
-    def explore_node_ouputs(self, node, all_paths, current_path):
-        '''
-        explore the nodes directlty down path from input node 
-        '''
-        print(node, "current path", current_path, all_paths)
-        if self.nodes[node].output_nodes:
-            for output_node in self.nodes[node].output_nodes:
-                current_path.append(output_node)
-                if output_node == 2: print(node, self.nodes[node].output_nodes)
+            if "Number" in entry: 
+                string = str(entry) + "\t" + str(stats[entry])
+                output = output  + "\n" + string 
+        return output
 
-                if self.nodes[output_node].node_type == "3_prime":
-                    print(current_path, all_paths)
-                    all_paths.append(current_path)
-                    current_path = []
-                    return all_paths
-                
-                else:
-                    # print("exploring: ", output_node, self.nodes[output_node].output_nodes)
-                    self.explore_node_ouputs(output_node, all_paths, current_path)
-        else:
-            print(node, current_path)
 
 
 
@@ -494,8 +640,7 @@ class Node(object):
         self.output_edges = edges_out
         self.input_nodes = nodes_in 
         self.output_nodes = nodes_out 
-        self.node_start = coordinates[0]
-        self.node_stop = coordinates[1]
+        self.node_start = coordinates
         self.frame = self.node_start % 3
 
 
@@ -542,33 +687,12 @@ class Edge(object):
 
 
 
-if __name__ == "__main__":
+# if __name__ == "__main__":
 
-    dg = RDG()
-    dg = dg.load_example()
-    node_key = dg.get_new_node_key()
-
-    start_codon_position = 15
-    stop_codon_position = 25
-    start_node = Node(key=node_key, node_type="start_codon", coordinates=(start_codon_position, start_codon_position + 2), edges_in=[], edges_out=[], nodes_in=[], nodes_out=[])
-
-    node_key = node_key + 1
-    stop_node = Node(key=node_key, node_type="stop_codon", coordinates=(stop_codon_position, stop_codon_position + 2), edges_in=[], edges_out=[], nodes_in=[], nodes_out=[])
-    
-    # dg.insert_ORF(dg.edges[2], start_node, stop_node)
-    print(dg.root_to_node_of_acyclic_edge_path(5))
-
-
-    # dg.add_open_reading_frame(30, 90)
-
-    # dg.add_open_reading_frame(150, 171)
-    # dg.print_paths()
-
-
-    # for i in dg.nodes:
-        # print(dg.nodes[i].key, dg.nodes[i].node_start, dg.nodes[i].node_stop, dg.nodes[i].node_type, dg.nodes[i].output_nodes)
-    # print ()
-
-    # for i in dg.edges:
-    #     print(dg.edges[i].key, dg.edges[i].coordinates, dg.edges[i].edge_type)
+#     dg = RDG()
+#     dg = dg.load_example()
+#     dg.add_open_reading_frame(30, 90)
 
+#     dg.add_open_reading_frame(150, 171)
+#     dg.add_stop_codon_readthrough(90, 120)
+#     print(dg.get_orfs())
