@@ -10,4 +10,4 @@ def test_basic_nodes():
 def test_root_to_node_acyclic_node_path():
     g = RDG()
     g = RDG.load_example(g)
-    assert g.root_to_node_of_acyclic(5) == [1, 3, 4, 5]
\ No newline at end of file
+    assert g.root_to_node_of_acyclic_node_path(5) == [1, 3, 4, 5]
\ No newline at end of file
