@@ -8,10 +8,10 @@ def test_update_edge_input_to_new_to_node():
 
     start_codon_position = 15
     stop_codon_position = 25
-    start_node = Node(key=node_key, node_type="start_codon", coordinates=(start_codon_position, start_codon_position + 2), edges_in=[], edges_out=[], nodes_in=[], nodes_out=[])
+    start_node = Node(key=node_key, node_type="start_codon", coordinates=start_codon_position, edges_in=[], edges_out=[], nodes_in=[], nodes_out=[])
 
     node_key = node_key + 1
-    stop_node = Node(key=node_key, node_type="stop_codon", coordinates=(stop_codon_position, stop_codon_position + 2), edges_in=[], edges_out=[], nodes_in=[], nodes_out=[])
+    stop_node = Node(key=node_key, node_type="stop_codon", coordinates=stop_codon_position, edges_in=[], edges_out=[], nodes_in=[], nodes_out=[])
     
     g.update_edge(2, 1, 5, (1,1000))
     tf = 2 in g.edges
